#if 0
gcc test.c -g -I/usr/include/nss3 -I/usr/include/nspr4 -lnss3 && valgrind ./a.out
exit $?
#endif

#include <nspr.h>
#include <nss.h>
#include <pk11pub.h>
#include <ssl.h>
#include <secerr.h>
#include <secmod.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

static char* nss_get_password(PK11SlotInfo *slot, PRBool retry, void *arg) {
  (void)slot; /* unused */
  if(retry || NULL == arg)
    return NULL;
  else
    return (char *)PORT_Strdup((char *)arg);
}

void test_auth(PK11SlotInfo *slot, const char *filename, const char *passwd) {
  printf("file=\"%s\"\t passwd=\"%s\"\t [%s]\n", filename, passwd,
      (SECSuccess == PK11_Authenticate(slot, PR_TRUE, (char *)passwd))
      ? "OK"
      : "failed");
}

PK11GenericObject* load_cert(const char *filename) {
  PK11SlotInfo *slot = PK11_FindSlotByName("PEM Token #1");
  if(!slot)
    return NULL;

  CK_ATTRIBUTE tpl[4];

  CK_OBJECT_CLASS objClass = CKO_CERTIFICATE;
  tpl[0].type = CKA_CLASS;
  tpl[0].pValue = &objClass;
  tpl[0].ulValueLen = sizeof(objClass);

  tpl[1].type = CKA_LABEL;
  tpl[1].pValue = (char *)filename;
  tpl[1].ulValueLen = strlen(filename) + 1;

  CK_BBOOL cktrue = CK_TRUE;
  tpl[2].type = CKA_TOKEN;
  tpl[2].pValue = &cktrue;
  tpl[2].ulValueLen = sizeof(cktrue);

  tpl[3].type = CKA_TRUST;
  tpl[3].pValue = &cktrue;
  tpl[3].ulValueLen = sizeof(cktrue);

  PK11GenericObject *obj = PK11_CreateGenericObject(slot, tpl, 3, PR_FALSE);
  if (!obj)
    return NULL;

  PK11_FreeSlot(slot);
  return obj;
}

bool load_key(SECMODModule *mod, const char *filename, const char *passwd) {
  PK11SlotInfo *slot = PK11_FindSlotByName("PEM Token #1");
  if(!slot)
    return false;

  CK_ATTRIBUTE tpl[3];

  CK_OBJECT_CLASS objClass = CKO_PRIVATE_KEY;
  tpl[0].type = CKA_CLASS;
  tpl[0].pValue = &objClass;
  tpl[0].ulValueLen = sizeof(objClass);

  tpl[1].type = CKA_LABEL;
  tpl[1].pValue = (char *)filename;
  tpl[1].ulValueLen = strlen(filename) + 1;

  CK_BBOOL cktrue = CK_TRUE;
  tpl[2].type = CKA_TOKEN;
  tpl[2].pValue = &cktrue;
  tpl[2].ulValueLen = sizeof(cktrue);

  PK11GenericObject *obj = PK11_CreateGenericObject(slot, tpl, 3, PR_FALSE);
  if (!obj)
    return false;

  SECMOD_WaitForAnyTokenEvent(mod, 0, 0);
  PK11_IsPresent(slot);

  test_auth(slot, filename, passwd);

  PK11_DestroyGenericObject(obj);
  PK11_FreeSlot(slot);
  return true;
}

bool test_key(const char *filename, const char *passwd) {
  SECMODModule* mod = SECMOD_LoadUserModule("library=libnsspem.so name=PEM", NULL, PR_FALSE);
  if(!mod || !mod->loaded)
    return false;

  PK11GenericObject *obj = load_cert("./test.pem");
  if(!obj)
    return false;

  if(!load_key(mod, filename, passwd))
    return false;

  PK11_DestroyGenericObject(obj);
  SECMOD_UnloadUserModule(mod);
  SECMOD_DestroyModule(mod);

  return true;
}

int main() {
  if (0 != NSS_NoDB_Init(NULL))
    return EXIT_FAILURE;

  PK11_SetPasswordFunc(nss_get_password);

  if(!test_key("key-err.pem", "A!m(89$mC"))
    return EXIT_FAILURE;

  if(!test_key("key-ok.pem", "asDF89*("))
    return EXIT_FAILURE;

  NSS_Shutdown();

  return EXIT_SUCCESS;
}

